//-------//-------//-------//-------//-------//-------//-------//-------//-------
//
// karp.cpp : TEST karplus strong VST.
//
//-------//-------//-------//-------//-------//-------//-------//-------//-------
//-------//-------//-------//-------//-------//-------//-------//-------//-------
// inc
//-------//-------//-------//-------//-------//-------//-------//-------//-------
#include "audioeffectx.h"
#include <math.h>

//-------//-------//-------//-------//-------//-------//-------//-------//-------
// def
//-------//-------//-------//-------//-------//-------//-------//-------//-------
typedef float t_snd;

enum {
  Bits         = (sizeof(t_snd) * 8),
  Channel      = (2),
  Freq         = (44100),
  Align        = ((Channel * Bits) / 8),
  BytePerSec   = (Freq * Align),
  BufNum       = 6,
  Samples      = 512,
};

//-------//-------//-------//-------//-------//-------//-------//-------//-------
// random
//-------//-------//-------//-------//-------//-------//-------//-------//-------
int random()
{
  static int a = 1;
  static int b = 234567;
  static int c = 890123;
  a += b;
  b += c;
  c += a;
  return (a >> 16);
}

float frand()
{
  return (float)random() / float(0x7FFF);
}


//-------//-------//-------//-------//-------//-------//-------//-------//-------
//
// karplus strong osc
//
//-------//-------//-------//-------//-------//-------//-------//-------//-------
struct Osc {
  #define LenMax  (Freq * 2)
  int Idx, Len, Inter;
  t_snd buf0[LenMax];
  
  void On(int note, int it = 2) {
    Len = LenMax / (int)pow(2.0, double(note) / 12.00);
    for(int i = 0 ; i < Len; i++)
      buf0[i] = float(random()) / 32768.0;
    Inter = it;
  }

  t_snd Get() {
    t_snd g = 0;
    if(Len <= 0) return 0;
    Idx++;
    
    //Interppp
    for(int i = 0 ; i < Inter; i++)
      g += buf0[(Idx + i) % Len] * 0.997;
    g /= Inter;
    buf0[Idx % Len] = g;
    return g;
  }
};

//-------//-------//-------//-------//-------//-------//-------//-------//-------
// Delay
//-------//-------//-------//-------//-------//-------//-------//-------//-------
struct DelayBuffer
{
  enum
  {
    Max = 100000,
  };
  float Buf[Max];
  unsigned long   Rate;
  unsigned int   Index;
  void Init(unsigned long m)        { Rate = m;  }
  void Update(float a)              { Buf[Index++ % Rate ] = a;  }
  float Sample(unsigned long n = 0) { return Buf[ (Index + n) % Rate]; }
};

//-------//-------//-------//-------//-------//-------//-------//-------//-------
// Reverb
// http://www.ari-web.com/service/soft/reverb-2.htm
//-------//-------//-------//-------//-------//-------//-------//-------//-------
struct Reverb
{
  enum
  {
    CombMax = 4,
    AllMax  = 2,
  };
  DelayBuffer comb[CombMax];
  DelayBuffer all[AllMax];

  float Sample(float a, int index = 0, int character = 0, int lpfnum = 4)
  {
    const int tau[][4][4] = 
    {
      //prime
      {
        {2063, 1847, 1523, 1277}, {3089, 2927, 2801, 2111}, {5479, 5077, 4987, 4057}, {9929, 7411, 4951, 1063},
      },
      {
        {2053, 1867, 1531, 1259}, {3109, 2939, 2803, 2113}, {5477, 5059, 4993, 4051}, {9949, 7393, 4957, 1097},
      }
    };

    const float gain[] = 
    {
       -(0.8733), -(0.8223), -(0.8513), -(0.8503),
    };
    float D = a * 0.5;
    float E = 0;
    
    //Comb
    for(int i = 0 ; i < CombMax; i++)
    {
      DelayBuffer *reb = &comb[i];
      reb->Init(tau[character % 2][index % 4][i]);
      float k = 0;
      float c = 0;
      int LerpMax = lpfnum + 1;
      for(int h = 0 ; h < LerpMax; h++)
        k += reb->Sample(h * 2);
      k /= float(LerpMax);
      c = a + k;
      reb->Update(c * gain[i]);
      E += c;
    }
    D = (D + E) * 0.3;
    return D;
  }
};

//------------//------------//------------//------------//------------//------------
//
// VST : karp
//
//------------//------------//------------//------------//------------//------------
class Karp : public AudioEffectX
{
  #define OSC_MAX 16
  Osc osc[OSC_MAX];
  Reverb rebL;
  Reverb rebR;

public:

  Karp (audioMasterCallback audioMaster)
  : AudioEffectX (audioMaster, 1, 0) //preset:1, parameter:0
  {
    setNumInputs(2);
    setNumOutputs(2);
    setUniqueID('ZZ15');
    isSynth(true);
    canProcessReplacing();
  }

  virtual void processReplacing (float** inputs, float** outputs, VstInt32 sampleFrames)
  {
    float *left  = outputs[0];
    float *right = outputs[1];
    
    for(int i = 0; i < sampleFrames; i++)
    {
      t_snd v = 0;
      for(int o = 0 ; o < OSC_MAX; o++)
        v += osc[o].Get();

      //koko kara tekitou
      static t_snd mult[] = {
        0.3, -0.5, 0.7
      };
      static t_snd z[3] = {0};
      for(int tap = 0; tap < 3; tap++) {
        v = v + (z[tap] - v) * mult[tap];
        z[tap] = v;
      }
      left [i] = 0.005 * v + (rebL.Sample(v, 0, 0, 2));
      right[i] = 0.005 * v + (rebR.Sample(v, 0, 1, 2));
    }
  }
  
  VstInt32 processEvents (VstEvents* events)
  {
    for(int i = 0 ; i < events->numEvents;i++)
    {
      if((events->events[i])->type == kVstMidiType)
      {
        VstMidiEvent *midievent = (VstMidiEvent*)(events->events[i]);
        int ev   = midievent->midiData[0] & 0xF0;
        int ch   = midievent->midiData[0] & 0x0F;
        int note = midievent->midiData[1] & 0x7F;
        
        //http://www.g200kg.com/jp/docs/tech/midi.html
        //L[on
        if(ev == 0x90)
        {
          osc[ch % OSC_MAX].On(note, 6);
        }
      }
    }
    return 1;
  }
};

AudioEffect* createEffectInstance (audioMasterCallback audioMaster)
{
  return new Karp(audioMaster);
}

