//-------//-------//-------//-------//-------//-------//-------//-------//-------
//
// karp.cpp : TEST simple VST
//
//-------//-------//-------//-------//-------//-------//-------//-------//-------
//-------//-------//-------//-------//-------//-------//-------//-------//-------
// inc
//-------//-------//-------//-------//-------//-------//-------//-------//-------
#include "audioeffectx.h"
#include <math.h>

//-------//-------//-------//-------//-------//-------//-------//-------//-------
// def
//-------//-------//-------//-------//-------//-------//-------//-------//-------
typedef float t_snd;


float rn()
{
  static int a = 1;
  static int b = 234567;
  static int c = 890123;
  a += b;
  b += c;
  c += a;
  return (float)(a >> 16) / float(0x7FFF);
}

//-------//-------//-------//-------//-------//-------//-------//-------//-------
//
// simple detune synth
//
//-------//-------//-------//-------//-------//-------//-------//-------//-------
struct Osc {
  double ph, fq, vol;
  Osc()
  {
    vol = ph = fq = 0;
  }
  void On(int note, int it = 2) {
    ph = 0;
    fq = 1.0 / 44100.0;
    for(int i = 0; i < note; i++)
      fq *= 1.0594630943592952645618252949463;
    vol = 1;
  }

  t_snd Get() {
    double k = ph;
    t_snd v = 0;
    if(vol <= 0.00001) return 0;
    v += (-1 + 2 * (k - floor(k))) > 0 ? 1 : -1;
    k *= 1.005;
    v += (-1 + 2 * (k - floor(k)));
    
    v /= 2;
    vol *= 0.99995;
    ph  += fq;
    return v * vol;
  }
};


//------------//------------//------------//------------//------------//------------
//
// VST : karp
//
//------------//------------//------------//------------//------------//------------
class Karp : public AudioEffectX
{
  #define OSC_MAX 16
  Osc osc[OSC_MAX];
public:

  Karp (audioMasterCallback audioMaster)
  : AudioEffectX (audioMaster, 1, 0) //preset:1, parameter:0
  {
    setNumInputs(2);
    setNumOutputs(2);
    setUniqueID('AZ16');
    isSynth(true);
    canProcessReplacing();
  }

  virtual void processReplacing (float** inputs, float** outputs, VstInt32 sampleFrames)
  {
    float *left  = outputs[0];
    float *right = outputs[1];
    
    for(int i = 0; i < sampleFrames; i++)
    {
      t_snd v = 0;
      for(int o = 0 ; o < OSC_MAX; o++)
        v += osc[o].Get();

      //tekitou
      left [i] = 0.5 * v;
      right[i] = 0.5 * v;
    }
  }
  
  VstInt32 processEvents (VstEvents* events)
  {
    for(int i = 0 ; i < events->numEvents;i++)
    {
      if((events->events[i])->type == kVstMidiType)
      {
        VstMidiEvent *midievent = (VstMidiEvent*)(events->events[i]);
        int ev   = midievent->midiData[0] & 0xF0;
        int ch   = midievent->midiData[0] & 0x0F;
        int note = midievent->midiData[1] & 0x7F;
        
        //http://www.g200kg.com/jp/docs/tech/midi.html
        //L[on
        if(ev == 0x90)
        {
          osc[ch % OSC_MAX].On(note, 6);
        }
      }
    }
    return 1;
  }
};

AudioEffect* createEffectInstance (audioMasterCallback audioMaster)
{
  return new Karp(audioMaster);
}

